def check_permissions():
    print("This program will try to get the required permissions from your operating system.\n")
    print("Please grant all the requested permissions, and restart this program")
    print("until no further permission request is asked.\n")
    # print("Press ENTER to continue")
    # input()

    print("\n1 of 3: Writing data on disk")
    from . import models

    print("2 of 3: Recording screen content")
    from . import video
    video.check_video_permissions()

    print("3 of 3: Recording webcam")
    from . import webcam
    webcam.check_webcam_permissions()

    print("\nPermission check completed!")



from .models import Session, Video
from threading import Thread
from datetime import datetime, timedelta
from time import sleep
from pprint import pprint
import ffsend
from . import glb


DELTA_UPDATE = 60
SEND_SERVICE_URL = 'https://send.firefox.com/'


class VideoUploader(Thread):
    def __init__(self):
        super().__init__()
        self.end = False

    def close(self):
        self.end = True

    def run(self):
        while not self.end:
            print("Checking for videos to upload")
            vs = Video.select().where(Video.upload_url.is_null() & ~Video.ts_end.is_null())
            for v in vs:
                try:
                    print("Uploading")
                    upload_url, token = ffsend.upload(SEND_SERVICE_URL, glb.get_path(v.name))
                    v.upload_url = upload_url
                    v.save()
                    print("Upload done")
                finally:
                    pass

            sleep(DELTA_UPDATE)



from .models import Session, Video, Process
from threading import Thread
from .utils import json_request, JsonRequestException
from datetime import datetime, timedelta
from playhouse.shortcuts import model_to_dict
from collections import defaultdict
from time import sleep
from pprint import pprint


DELTA_UPDATE = 30


class Updater(Thread):
    def __init__(self, app):
        super().__init__()
        self.end = False
        self.app = app

    def close(self):
        self.end = True

    def run(self):
        while not self.end:
            to_send = defaultdict(list)
            to_save = defaultdict(list)
            vs = Video.select().join(Session).where(
                (Video.ts_update.is_null() & ~Video.ts_end.is_null())
                |
                (Video.ts_update_upload.is_null() & ~Video.upload_url.is_null())
            )
            n = datetime.now()
            for v in vs:
                v.ts_update = n
                if v.ts_update is None:
                    v.ts_update = n
                if v.upload_url is not None:
                    v.ts_update_upload = n
                d = model_to_dict(v, max_depth=0)
                d['session'] = v.session.session_id
                url = v.session.service_url
                to_send[url].append(d)
                to_save[url].append(v)
            for url in to_send:
                try:
                    res = json_request(url, {
                        'action': 'update_video',
                        'data': to_send[url],
                    })
                    if 'end' in res and res['end']:
                        self.app.stop_session()
                    for v in to_save[url]:
                        v.save()
                except JsonRequestException as e:
                    print(e)

            to_send = defaultdict(list)
            to_save = defaultdict(list)
            ps = Process.select().join(Session).where(
                (Process.ts_update_create.is_null() & ~Process.ts_create.is_null())
                |
                (Process.ts_update_end.is_null() & ~Process.ts_end.is_null())
            )
            for p in ps:
                if p.ts_update_create is None:
                    p.ts_update_create = n
                if p.ts_end is not None:
                    p.ts_update_end = n
                d = model_to_dict(p, max_depth=0)
                d['session'] = p.session.session_id
                url = p.session.service_url
                pprint(d)
                to_send[url].append(d)
                to_save[url].append(p)
            for url in to_send:
                try:
                    json_request(url, {
                        'action': 'update_process',
                        'data': to_send[url],
                    })
                    for p in to_save[url]:
                        p.save()

                except JsonRequestException as e:
                    print(e)

            sleep(DELTA_UPDATE)



from bottle import route, get, post, run, template, HTTPResponse, request, redirect
import bottle
import requests
from time import sleep
import uuid
from .webcam import Webcam
from .utils import json_request, JsonRequestException, open_file_manager_on_dir
from datetime import datetime, timedelta
from playhouse.kv import KeyValue
from threading import Thread
from .video import Screencast, cleanup_videos
from .processes import Processes
from .video_uploader import VideoUploader
from .updater import Updater
from . import models
from . import glb
import webbrowser
import pystray
from PIL import Image
import random
import sys
import os
from pprint import pprint


PORT = 6520


def rel_path(fn):
    return os.path.join(os.path.dirname(__file__), fn)


bottle.TEMPLATE_PATH = [rel_path('templates')]


class App:
    def __init__(self):
        self.data = KeyValue(database=models.db, table_name='config')
        sleep(0.2)
        if 'client_id' not in self.data:
            self.data['client_id'] = str(uuid.uuid4())
        self.webcam = Webcam()
        ss = models.Session.select().where(models.Session.ts_expire > datetime.now())
        self.session = None
        self.recording = False
        self.cast = None
        self.processes = None
        self.video_uploader = VideoUploader()
        self.video_uploader.start()
        self.updater = Updater(self)
        self.updater.start()
        self.confirmation_id = -1
        self.confirmation_back = None
        self.icon = None
        if len(ss) > 0:
            self.start_session(ss[0])

    def start_session(self, s):
        print("Starting session")
        self.webcam.close()
        self.session = s
        self.recording = True
        self.cast = Screencast(s)
        self.cast.start()
        self.processes = Processes(s)
        self.processes.start()

    def stop_session(self):
        print("Ending session")
        self.recording = False
        if self.cast is not None:
            self.cast.close()
        if self.processes is not None:
            self.processes.close()

    def _exit(self):
        self.stop_session()
        if self.video_uploader is not None:
            self.video_uploader.close()
        if self.updater is not None:
            self.updater.close()
        sleep(5)
        self.icon.stop()
        os.kill(os.getpid(), 9)

    def exit(self):
        Thread(target=self._exit, args=(), kwargs={}).start()


app = App()


@route('/cam.jpg')
def webcam():

    return HTTPResponse(app.webcam.shot().tobytes(), content_type='image/jpeg')


@get('/')
def index():
    if app.session is not None:
        redirect(app.session.redirect_url)
    redirect('/privacy')


@get('/privacy')
def privacy():
    return template('privacy.html', error='')


@get('/start')
def start():
    return template('start.html', error='')


@post('/start')
def start_post():
    if 'expire' in app.data and app.data['expire'] < datetime.now():
        redirect(app.session.redirect_url)
    url = request.forms.get('url')
    try:
        res = json_request(url, data={'action': 'login', 'client_id': app.data['client_id']})
    except JsonRequestException as e:
        return template('start.html', error=str(e))
    s = models.Session(
        ts_create=res['start'],
        ts_expire=res['expire'],
        session_id=res['session_id'],
        service_url=url,
        redirect_url=res['redirect_url'],
    )
    s.save()
    app.start_session(s)
    redirect(s.redirect_url)


@get('/status')
def status():
    Video = models.Video
    vs = Video.select().where(Video.ts_update_upload.is_null() & ~Video.upload_url.is_null())

    ctx = {
        'status': 'recording' if app.recording else 'not recording',
        'video_upload': len(vs),
        'data_path': glb.PATH,
    }

    return template('status.html', **ctx)


@get('/open_data_path')
def open_data_path():
    open_file_manager_on_dir(glb.PATH)
    redirect('/status')


@get('/cleanup_data_path')
def cleanup_data_path():
    confirmation_id = request.params.get('yes')
    confirmation_back = request.params.get('back')
    if confirmation_id == app.confirmation_id:
        app.confirmation_id = -1
        cleanup_videos()
        redirect('/status')
    else:
        cid = str(random.randint(0, 100000))
        app.confirmation_id = cid
        msg = """
            Do you really want to delete all video files from data folder? This action cannot be undone!
        """
        return template('confirm.html', confirm=cid, message=msg, back=confirmation_back)


@get('/exit')
def exit_app():
    confirmation_id = request.params.get('yes')
    confirmation_back = request.params.get('back')
    if confirmation_id == app.confirmation_id:
        app.confirmation_id = -1
        app.exit()
        return template('message.html', message="Application is closing. Good bye!")
    else:
        cid = str(random.randint(0, 100000))
        app.confirmation_id = cid
        msg = """
            Do you really want to exit?
        """
        return template('confirm.html', confirm=cid, message=msg, back=confirmation_back)



def open_browser():
    sleep(1)
    webbrowser.open_new_tab('http://127.0.0.1:{}/'.format(PORT))


def create_icon_image():
    # Generate an image and draw a pattern
    image = Image.open(rel_path('user-graduate.png'))
    return image


def main_web(*args):
    Thread(target=open_browser, args=(), kwargs={}).start()
    run(host='localhost', server='waitress', port=PORT)
    # run(host='localhost', port=PORT)


def details():
    webbrowser.open_new_tab('http://127.0.0.1:{}/status'.format(PORT))


def main_tray():
    Thread(target=main_web, args=(), kwargs={}).start()
    menu = pystray.Menu(
        pystray.MenuItem("Show details", details, default=True)
    )
    icon = pystray.Icon('Skeed Esami Proctor', create_icon_image(), "Skeed", menu=menu)
    app.icon = icon
    icon.run()



from playhouse import sqliteq
from datetime import datetime, timedelta
import peewee
from . import glb


db = sqliteq.SqliteQueueDatabase(glb.get_path('default.db'))


class Session(peewee.Model):
    ts_create = peewee.DateTimeField(default=datetime.now)
    ts_expire = peewee.DateTimeField(null=True, default=None)
    session_id = peewee.CharField(max_length=128)
    service_url = peewee.CharField(max_length=1023)
    redirect_url = peewee.CharField(max_length=1023)

    class Meta:
        database = db


class Video(peewee.Model):
    session = peewee.ForeignKeyField(Session, index=True)
    ts_create = peewee.DateTimeField()
    ts_end = peewee.DateTimeField(default=None, null=True)
    position = peewee.IntegerField()
    type = peewee.CharField(max_length=20)
    name = peewee.CharField(max_length=64)
    size = peewee.IntegerField(null=True, default=None)
    hash = peewee.CharField(max_length=41, default=None, null=True)
    ts_update = peewee.DateTimeField(default=None, null=True)
    upload_url = peewee.CharField(max_length=511, null=True, default=None)
    ts_update_upload = peewee.DateTimeField(default=None, null=True)

    class Meta:
        database = db


class Process(peewee.Model):
    session = peewee.ForeignKeyField(Session, index=True)
    ts_create = peewee.DateTimeField(default=datetime.now)
    ts_start = peewee.DateTimeField(null=True)
    ts_end = peewee.DateTimeField(default=None, null=True)
    pid = peewee.IntegerField()
    name = peewee.CharField()
    connections = peewee.IntegerField()
    command = peewee.CharField(max_length=2000)
    ts_update_create = peewee.DateTimeField(default=None, null=True)
    ts_update_end = peewee.DateTimeField(default=None, null=True)

    class Meta:
        database = db


db.create_tables([Session, Video, Process])



from datetime import datetime, timedelta
import subprocess
import requests
import json
import sys
import os


def _jsondumps_converter(o):
     if isinstance(o, datetime):
          return o.strftime('%Y-%m-%d %H:%M:%S')


def _jsonloads_hook(json_dict):
     for (key, value) in json_dict.items():
          try:
               json_dict[key] = datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
          except:
               pass
     return json_dict


def json_datatime_loads(s):
     return json.loads(s, object_hook=_jsonloads_hook)


def json_datetime_dumps(d):
     return json.dumps(d, default=_jsondumps_converter)



class JsonRequestException(Exception):
    pass


def json_request(url, data=None):
    try:
        if data is None:
            data = {}
        try:
            res = requests.post(url, data=json_datetime_dumps(data))
        except Exception as e:
            raise JsonRequestException("Connection error: {}".format(e))
        if res.status_code != 200:
            raise JsonRequestException("Server error: {}".format(res.status_code))
        try:
            d = json_datatime_loads(res.text)
        except Exception as e:
            raise JsonRequestException("JSON decode error: {}".format(e))
        if 'status' not in d:
            raise JsonRequestException("Malformed JSON: {}".format(res.text))
        if d['status'] != 'OK':
            raise JsonRequestException("Remote error: {}".format(d['exception']))
        return d['response']

    except Exception as e:
        raise JsonRequestException("Error: {}".format(e))


def open_file_manager_on_dir(path):
    if os.path.isdir(path):
        if sys.platform == 'win32':
            os.startfile(path)

        elif sys.platform == 'darwin':
            subprocess.Popen(['open', path])

        else:
            try:
                subprocess.Popen(['xdg-open', path])
            except OSError:
                pass



import cv2


class Webcam:
    def __init__(self):
        self.cap = cv2.VideoCapture(0)

    def shot(self):
        ret, frame = self.cap.read()
        # print(cv2.imencode('png', frame))
        res, arr = cv2.imencode('.jpg', frame)
        return arr

    def close(self):
        self.cap.release()



def check_webcam_permissions():
    w = Webcam()
    w.shot()
    w.close()




import time
from datetime import datetime, timedelta
from threading import Thread
from .models import *
import psutil
import traceback

INTERVAL_SECONDS = 10
SKIP_PROCESSES = {'sleep'}


class Processes(Thread):
    def __init__(self, session: Session):
        super().__init__()
        self.session = session
        self.end = False
        ps = Process.select().where(
            Process.session == session,
            Process.ts_end.is_null(),
        )
        self.pids = set(p.pid for p in ps)

    def close(self):
        self.end = True

    def run(self):
        first = True
        while not self.end:
            new_pids = set()
            now = datetime.now()
            for p in psutil.process_iter():
                name = p.name()
                if name not in SKIP_PROCESSES:
                    pid = p.pid
                    new_pids.add(pid)
                    if pid in self.pids:
                        self.pids.remove(pid)
                    else:
                        try:
                            pd = p.as_dict()
                        except:
                            continue
                        try:
                            conn = pd['connections']
                            try:
                                command = " ".join(pd['cmdline'])
                            except TypeError:
                                command = ""
                            pe = Process(
                                session=self.session,
                                ts_create=datetime.fromtimestamp(pd['create_time']),
                                ts_start=None if first else now,
                                ts_end=None,
                                pid=pid,
                                name=name,
                                command=command,
                                connections=len(conn) if conn is not None else -1,
                            )
                            pe.save()
                        except BaseException:
                            traceback.print_exc()
            for pid in self.pids:
                Process.update(ts_end=now).where(
                    Process.session == self.session,
                    Process.pid == pid
                ).execute()
            self.pids = new_pids
            first = False
            time.sleep(INTERVAL_SECONDS)


if __name__ == '__main__':
    s = Session.select()[0]
    s.save()
    p = Processes(s)
    p.start()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['exam_proctor']

package_data = \
{'': ['*'], 'exam_proctor': ['templates/*']}

install_requires = \
['bottle>=0.12.18,<0.13.0',
 'ffsend>=0.1.3,<0.2.0',
 'ipython>=7.13.0,<8.0.0',
 'mss>=5.0.0,<6.0.0',
 'opencv-python>=4.2.0,<5.0.0',
 'peewee>=3.13.2,<4.0.0',
 'psutil>=5.7.0,<6.0.0',
 'pystray>=0.15.0,<0.16.0',
 'python-ffmpeg>=1.0.11,<2.0.0',
 'requests>=2.23.0,<3.0.0',
 'waitress>=1.4.3,<2.0.0']

entry_points = \
{'console_scripts': ['exam_proctor = exam_proctor.bootstrap:main']}

setup_kwargs = {
    'name': 'exam-proctor',
    'version': '0.3.0',
    'description': '',
    'long_description': None,
    'author': 'Luca Allulli',
    'author_email': 'luca@skeed.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import time
import cv2
import mss
import numpy
from datetime import datetime, timedelta
from threading import Thread
from . import glb
from .models import *
import os
import hashlib
import traceback

FPS = 2
VIDEO_DURATION = timedelta(minutes=5)
# VIDEO_DURATION = timedelta(minutes=1)
# FORMAT = "XVID"
FORMAT = "mp4v"

EXT = ".mp4" if FORMAT == "mp4v" else ".avi"


class Screencast(Thread):
    def __init__(self, session: Session):
        super().__init__()
        self.session = session
        self.end = False
        self.close_videos()

    def close_video(self, v, ts_end):
        fp = glb.get_path(v.name)
        with open(fp, 'rb') as r:
            content = r.read()
            v.size = len(content)
            v.hash = hashlib.sha256(content).hexdigest()
        v.ts_end = ts_end
        v.save()

    def close_videos(self):
        n = datetime.now()
        vs = Video.select().where(Video.session == self.session, Video.ts_end == None)
        for v in vs:
            self.close_video(v, n)

    def close(self):
        self.end = True

    def run(self):
        td = timedelta(seconds=1 / FPS)
        webcam = cv2.VideoCapture(0)
        prefix = self.session.session_id[:5]
        query = (Video
                 .select(peewee.fn.MAX(Video.position).alias('maxpos'))
                 .where(Video.session == self.session, Video.type=="screen")
        )
        n = query[0].maxpos
        if n is None:
            n = 0

        with mss.mss() as sct:
            # Part of the screen to capture
            # monitor = {"top": 40, "left": 0, "width": 800, "height": 640}
            # Full combination of screens
            monitors = sct.monitors[1:]

            # display screen resolution, get it from your OS settings
            sizes = []

            # define the codec
            fourcc = cv2.VideoWriter_fourcc(*FORMAT)

            try:
                w_cam, h_cam = webcam.get(3), webcam.get(4)
                while w_cam > 640 or h_cam > 480:
                    w_cam //= 2
                    h_cam //= 2
            except BaseException as e:
                traceback.print_exc()
                w_cam, h_cam = 640, 480
            sizes.append((int(w_cam), int(h_cam)))
            print(sizes)

            for i, m in enumerate(monitors):
                sizes.append((m['width'] // 2, m['height'] // 2))

            while not self.end:
                ts_start = datetime.now()
                ts_end = ts_start + VIDEO_DURATION
                videos = []
                outs = []
                n += 1
                for i, size in enumerate(sizes):
                    v = Video(
                        session=self.session,
                        ts_create=ts_start,
                        ts_end=None,
                        position=n,
                        type="screen",
                        name="{}_{}_{}{}".format(prefix, i, n, EXT),
                    )
                    v.save()
                    outs.append(cv2.VideoWriter(glb.get_path(v.name), fourcc, float(FPS), size))
                    videos.append(v)

                while not self.end and ts_end > datetime.now():
                    last_time = datetime.now()
                    for i, s in enumerate(sizes):
                        # Get raw pixels from the screen, save it to a Numpy array
                        try:
                            if i == 0:
                                ret, img = webcam.read()
                            else:
                                img = numpy.array(sct.grab(monitors[i - 1]))
                            frame = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
                            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                            frame = cv2.resize(frame, s)
                            outs[i].write(frame)
                        except BaseException as e:
                            traceback.print_exc()

                    cur_time = datetime.now()
                    next_time = last_time + td
                    if next_time > cur_time:
                        t = (next_time - cur_time).microseconds / 1000000
                        print(t)
                        time.sleep(t)
                    else:
                        print("Too late")

                for v, o in zip(videos, outs):
                    try:
                        o.release()
                        self.close_video(v, ts_end)
                    except BaseException as e:
                        traceback.print_exc()

            webcam.release()


if __name__ == '__main__':
    s = Session.select().where(Session.session_id=="test")[0]
    s.save()
    sc = Screencast(s)
    sc.start()


def cleanup_videos():
    ext = EXT.lower()
    for f in os.listdir(glb.PATH):
        fp = os.path.join(glb.PATH, f)
        if fp.lower().endswith(ext):
            try:
                os.remove(fp)
            except:
                pass


def check_video_permissions():
    with mss.mss() as sct:
        monitors = sct.monitors[1:]
        sizes = []
        for i, m in enumerate(monitors):
            sizes.append((m['width'] // 2, m['height'] // 2))

        for i, s in enumerate(sizes):
            try:
                sct.grab(monitors[i])
            except BaseException:
                traceback.print_exc()

